/*==== T a p T e m p o C l o c k . c ==================================================================================
    
    Copyright (c) 2014 Harald Sabro. For non-commercial use only. Do not post other than at www.sabrotone.com.
    
    Project     : Tap-Tempo Clock using the ATtiny861
    Author      : Harald Sabro
    
    Note        : Debounce code based on article at: http://www.ganssle.com/debouncing-pt2.htm
                  
                  PWM DDS implementation inspired by:
                  http://interface.khm.de/index.php/lab/experiments/arduino-dds-sinewave-generator/
                  
                  Syncing on first tap, and selecting options via voltage divider/pot; ideas (not code) borrowed from:
                  http://www.electricdruid.net/index.php?page=projects.taplfo
                  
    Fuse bits   : Clear "divide clock by 8".
                  Set clock source to external 8MHz+ crystal oscillator.
    
 ====================================================================================================================*/ 

/*====== Includes =====================================================================================================
=====================================================================================================================*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>

/*====== Defines ======================================================================================================
=====================================================================================================================*/

///////////////////////////////////////////////////////////////////////////////////////////////////
// Couldn't find where the bool_t was defined so here it is.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define bool_t      uint8_t
#define false       0
#define true        1

///////////////////////////////////////////////////////////////////////////////////////////////////
// Defines related to timer setup.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define CLOCK_FREQUENCY             8000000UL   // Fuse bits set to external 8MHz crystal.

#define PWM_OUTPUT_RESOLUTION       256

#define TIMER0_PRE_SCALER           64.0
#define TIMER0_FREQUENCY            0.001       // Timer0 trigger rate in Hertz (every 1ms).
#define SWITCH_SAMPLER_RATE         1           // Switch sampler rate in milliseconds.

#define TIMER1_SAMPLE_RATE          (CLOCK_FREQUENCY / PWM_OUTPUT_RESOLUTION)   // 8Mhz / 256 = 31.25kHz

///////////////////////////////////////////////////////////////////////////////////////////////////
// Defines related to the key debounce logic.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define DEBOUNCE_CHECK_COUNT		10          // Millisecond count before considering the switch state stable.

#define DEFAULT_TEMPO               1000        // In milliseconds = 1Hz

///////////////////////////////////////////////////////////////////////////////////////////////////
// Minimum and maximum range for tap tempo, not counting in any multiplier.
//
// Note: An upper tap-tempo frequency of 20Hz coincides nicely with a low pass filter of
//       6k8 / 1uF ~= 23Hz to get rid of any PWM high-frequency noise.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define LFO_MIN_TAP_TEMPO           10000       // In milliseconds = 0.1Hz, 10 seconds
#define LFO_MAX_TAP_TEMPO           50          // In milliseconds = 20Hz, 5/100ths of a second

#define ADJUSTED_MIN_TEMPO          10000.0     // Same as LFO_MIN_TAP_TEMPO
#define ADJUSTED_MAX_TEMPO          12.5        // LFO_MAX_TAP_TEMPO / 4

///////////////////////////////////////////////////////////////////////////////////////////////////
// Defines related to the input selection pins and other enable/disable pins.
//
// Note: The enable/disable pins are purposefully defined so that the default state coincides with
//       the internal pull-up resistor so that the pin can be left unattached if a particular
//       feature isn't required.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define CLOCK_INPUT_SOURCE_ENABLED          0
#define TAP_INPUT_SOURCE_ENABLED            1

#define CONFIG_PIN_ENABLED                  0
#define CONFIG_PIN_DISABLED                 1

#define TEMPO_AVERAGE_MAX_COUNT             10

///////////////////////////////////////////////////////////////////////////////////////////////////
// Defines related to tempo adjustments.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define TEMPO_ADJUST_CENTER                 128

#define TEMPO_ADJUST_MIN_MS_INTERVAL        1
#define TEMPO_ADJUST_MAX_MS_INTERVAL        10

#define TEMPO_ADJUST_MULTIPLIER_DEFAULT     3.0

#define TEMPO_ADJUST_RANGE_DIVIDER          (256 / TEMPO_ADJUST_MAX_MS_INTERVAL)

///////////////////////////////////////////////////////////////////////////////////////////////////
// Defines making it easier to change pin configurations.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define DATA_REG_A              PORTA
#define DATA_REG_B              PORTB
#define DATA_DIR_REG_A          DDRA
#define DATA_DIR_REG_B          DDRB
#define INPUT_PINS_A            PINA
#define INPUT_PINS_B            PINB

#define CLOCK_OUT_PIN           PB0         // Clock output.
#define CLOCK_OUT_DBL_PIN       PB1         // 2x frequency clock output.
#define TAP_INPUT_ACTIVE_PIN    PB2         // LED indicator when active tempo input is enabled?
#define UNUSED1                 PB3
#define CRYSTAL_PIN1            PB4         // Crystal leg #1.
#define CRYSTAL_PIN2            PB5         // Crystal leg #2.
#define TAP_AVERAGING_PIN       PB6         // Tap switch inputs will accumulate and an average tempo calculated.
#define UNUSED2                 PB7
#define TAP_INPUT_PIN           PA0         // Tap switch input.
#define CLOCK_INPUT_PIN         PA1         // Clock sync input.
#define SYNC_INPUT_PIN          PA2         // Single sync switch input.
#define ADJUST_ENABLED_PIN      PA3         // Tempo adjust enable.
#define INPUT_SRC_PIN           PA4         // Input source selection (tap or clock).
#define CLOCK_SRC_IS_DBL_PIN    PA5         // The input clock source is twice the speed.
#define TEMPO_ADJUST_PIN        PA6         // Tempo adjust ADC input.
#define ADJUST_RANGE_PIN        PA7         // Adjust range ADC input.

/*====== Local function prototypes ====================================================================================
=====================================================================================================================*/

void DebounceSwitches();

void SetBaseTempo(uint16_t milliseconds);

void RecalculateTempo();

uint16_t CalculateAverageTempo(uint16_t base_tempo);

void AlignOutputClockSignal();// __attribute__((always_inline));

void StartTempoCount();

void ExitTempoCount();

/*====== Global variables =============================================================================================
=====================================================================================================================*/

volatile uint8_t g_debounced_state = 0xff;
volatile uint8_t g_switch_state_changed = 0x00;

volatile uint8_t g_clock_input = 0;
volatile uint8_t g_input_selection = TAP_INPUT_SOURCE_ENABLED;
volatile uint8_t g_tempo_adjustment_is_active = CONFIG_PIN_DISABLED;
volatile uint8_t g_input_clock_is_double_tempo = CONFIG_PIN_DISABLED;
volatile uint8_t g_tap_tempo_averaging = CONFIG_PIN_DISABLED;

volatile bool_t g_is_counting_double_tempo = false;

volatile uint8_t g_tempo_adjustment = TEMPO_ADJUST_CENTER;
volatile uint8_t g_adjustment_range = 0;
volatile double g_adjustment_multiplier = TEMPO_ADJUST_MULTIPLIER_DEFAULT;

volatile uint32_t g_phase_accumulator = 0;
volatile uint32_t g_duty_cycle = 0;
volatile uint8_t g_table_index = 0xff;

volatile bool_t g_is_counting_tempo = false;
volatile uint16_t g_tempo_count = 0;
volatile uint16_t g_base_tempo = 0;
volatile uint16_t g_average_tempo[TEMPO_AVERAGE_MAX_COUNT] = {0};
volatile uint8_t g_average_tempo_count = 0;

/*====== m a i n ( ) ==================================================================================================
    Purpose     : Entry point and main loop.
    Author      : Harald Sabro
=====================================================================================================================*/

int main()
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Initialization.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Set the CLOCK_OUT_* pins and the tap input active pin as outputs, and the rest as inputs.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    DATA_DIR_REG_A = 0x00;
    DATA_DIR_REG_B = (1 << CLOCK_OUT_PIN) | (1 << CLOCK_OUT_DBL_PIN) | (1 << TAP_INPUT_ACTIVE_PIN);
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Enable pull-up resistors on input pins and drive output pins high.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    DATA_REG_A = 0xff;
    DATA_REG_B = 0xff;
    
    SetBaseTempo(DEFAULT_TEMPO);              // Start off with a 1Hz clock output frequency.
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Set up Timer0 to trigger an interrupt every 1ms. CTC mode and a prescaler of 64.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    TCCR0A = (1 << CTC0);                                                   // Clear Timer on Compare match.
    TCCR0B = (1 << CS01) | (1 << CS00);                                     // CK/64.
    TCNT0L = 0x00;                                                          // Starting at 0.
    OCR0A = ((CLOCK_FREQUENCY / TIMER0_PRE_SCALER) * TIMER0_FREQUENCY) - 1; // Trigger value.
    TIMSK = (1 << OCIE0A);                                                  // Timer0 interrupt enabled.
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Set up Timer1 in fast PWM mode with prescaler 2 and a non-inverted compare. 
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    TCCR1A = (1 << PWM1B) | (1 << COM1B1);      // Enable PWM-mode on OCR1B. Clear on compare match.
    TCCR1B = (1 << CS10);                       // No prescaler.
    TCCR1D = 0x00;                              // WGM11:10 = 00 enables Fast PWM.
    TIMSK |= (1 << TOIE1);                      // Timer1 overflow interrupt enabled.
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Set up ADC input pins for tempo adjust- and adjust range selection. VCC as reference
    // voltage, and left-aligned to get an 8-bit resolution. We're using a prescaler of 128 for an
    // interrupt frequency of 62.5kHz (within the required 50kHz-200kHz without being over the
    // top). We're going to use both PA6 and PA7 so we need to multiplex between these two inputs.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    ADMUX = (1 << ADLAR) | (1 << MUX2) | (1 << MUX0);       // Left-adjust, input channel = PA6.
    ADCSRA = (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);    // Prescaler division = 128.
    ADCSRA |= (1 << ADEN) | (1 << ADIE);                    // ADC enabled, ADC interrupt enabled.
    ADCSRA |= (1 << ADSC);                                  // ADC start the first conversion.
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Set up PCINT1, PCINT3, PCINT4 and PCINT5 (PA1, PA3, PA4 and PA5) for pin change interrupts.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    PCMSK0 = (1 << PCINT1) | (1 << PCINT3) | (1 << PCINT4) | (1 << PCINT5); // Set PA1, PA4 and PA5 as pin change interrupt sources.
    PCMSK1 = 0x00;
    GIMSK |= (1 << PCIE1);                                                  // Enable pin change interrupts.

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Read the input selection pin to determine the initial input source.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    g_input_selection = (INPUT_PINS_A & (1 << INPUT_SRC_PIN)) >> 4;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Read the current state of the double-clock, tempo adjust and tap averaging enabler-pins.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    g_tempo_adjustment_is_active = (INPUT_PINS_A & (1 << ADJUST_ENABLED_PIN)) >> 3;
    g_input_clock_is_double_tempo = (INPUT_PINS_A & (1 << CLOCK_SRC_IS_DBL_PIN)) >> 5;
    g_tap_tempo_averaging = (INPUT_PINS_B & (1 << TAP_AVERAGING_PIN)) >> 6;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Enable global interrupts. No interrupts will happen without this, regardless of individual
    // flags set elsewhere.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    sei();
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Main loop.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    while (1)
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Nothing here. Everything happens within the interrupt loops.
        ///////////////////////////////////////////////////////////////////////////////////////////
    }
}

/*====== D e b o u n c e S w i t c h e s ( ) ==========================================================================
    Purpose     : 
    Author      : Harald Sabro
=====================================================================================================================*/

void DebounceSwitches()
{
	static uint8_t state_index = 0;
	static uint8_t state[DEBOUNCE_CHECK_COUNT] = {0};
	uint8_t count;
	uint8_t accumulated_state = 0x00;
	uint8_t previous_state;
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// Get the current state of all PA pins and store them at whatever state index we're currently
	// at, making sure to reset the state index whenever we go past the max count.
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	state[state_index] = INPUT_PINS_A;
	
	state_index++;
	if (state_index >= DEBOUNCE_CHECK_COUNT)
	{
		state_index = 0;
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	// Build an accumulated switch state based on DEBOUNCE_CHECK_COUNT number of stored states by
	// AND'ing everything together. If a particular switch pin stayed "on" the entire time the
	// corresponding debounce state bit will be set, and likewise, as soon as one or more reading
	// on a particular pin reads "off" that debounce state bit will be cleared.
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	for (count = 0; count < DEBOUNCE_CHECK_COUNT; count++)
	{
		accumulated_state |= state[count];
	}
	
	previous_state = g_debounced_state;
	g_debounced_state = accumulated_state;
	g_switch_state_changed = g_debounced_state ^ previous_state;
}

/*====== S e t B a s e T e m p o ( ) ==================================================================================
    Purpose     : 
    Author      : Harald Sabro
=====================================================================================================================*/

void SetBaseTempo(uint16_t milliseconds)
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Do some boundary checking on the requested frequency. We're only going to accept frequencies
    // in the range of 0.1Hz - 20Hz.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if ((milliseconds > LFO_MIN_TAP_TEMPO) || (milliseconds < LFO_MAX_TAP_TEMPO))
    {
        return;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // If tempo averaging is enabled the current reading is passed on for further calculation and
    // will more than likely be modified somewhat.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    if ((g_input_selection == TAP_INPUT_SOURCE_ENABLED) && (g_tap_tempo_averaging == CONFIG_PIN_ENABLED))
    {
        milliseconds = CalculateAverageTempo(milliseconds);
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Potentially save some calculating when nothing has changed (would be typical when running
    // off an external clock pulse).
    //
    // Note: May have to give or take a few milliseconds for this to really be of any value, but
    //       doesn't really hurt either.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (g_base_tempo != milliseconds)
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Store the new base tempo and recalculate how that affects the actual output.
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        g_base_tempo = milliseconds;
        
        RecalculateTempo();
    }
}

/*====== R e c a l c u l a t e T e m p o ( ) ==========================================================================
    Purpose     : 
    Author      : Harald Sabro
=====================================================================================================================*/

void RecalculateTempo()
{
    double final_tempo = (double)g_base_tempo;
    double new_frequency;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // If tempo adjustment is enabled (and we're not running off an external clock signal) modify
    // the base tempo before proceeding.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    //
    //
    // TODO: Consider moving the input source check somewhere prior to getting here. Might save us
    //       doing some calculations unnecessarily.
    //
    //

    if ((g_tempo_adjustment_is_active == CONFIG_PIN_ENABLED) && (g_input_selection != CLOCK_INPUT_SOURCE_ENABLED))
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Add or subtract to the tempo a number of millisecond intervals based on how far away
        // from "center" the adjustment input currently is.
        // Since the "center" setting is 256 / 2 = 128, and given a millisecond interval of 10ms,
        // this will allow the tempo adjustment to modify the tempo +/- 1,28sec. That's quite a
        // bit.
        // Add to this the "range" or millisecond interval, a simple multiplier to the tempo
        // adjustment.
        ///////////////////////////////////////////////////////////////////////////////////////////

        final_tempo -= (g_tempo_adjustment - TEMPO_ADJUST_CENTER) * g_adjustment_multiplier;

        ///////////////////////////////////////////////////////////////////////////////////////////
        // Make sure to keep within limits.
        ///////////////////////////////////////////////////////////////////////////////////////////

        if (final_tempo < ADJUSTED_MAX_TEMPO)
        {
            final_tempo = ADJUSTED_MAX_TEMPO;
        }
        else if (final_tempo > ADJUSTED_MIN_TEMPO)
        {
            final_tempo = ADJUSTED_MIN_TEMPO;
        }
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Convert the base tempo from a millisecond amount to a frequency.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    new_frequency = (1000.0 / final_tempo);
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Recalculate the base duty cycle based on the new frequency.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    g_duty_cycle = new_frequency * 0xffffffff / TIMER1_SAMPLE_RATE;
}

/*====== C a l c u l a t e A v e r a g e T e m p o ( ) ================================================================
    Purpose     : 
    Author      : Harald Sabro
=====================================================================================================================*/

uint16_t CalculateAverageTempo(uint16_t tempo)
{
    static uint8_t index = 0;
    uint8_t i;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Store the current tempo reading in the correct array spot and update the reading count if
    // we haven't yet reached the maximum.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    if (g_average_tempo_count < TEMPO_AVERAGE_MAX_COUNT)
    {
        index = g_average_tempo_count;
        g_average_tempo_count++;
    }

    g_average_tempo[index] = tempo;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Update the array index making sure to wrap around the index if we reach the end. This way we
    // always overwrite the oldest recorded tempo.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    index++;
    if (index >= TEMPO_AVERAGE_MAX_COUNT)
    {
        index = 0;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Add up and divide by sample count to get a basic average tempo. Repurposing the input
    // argument variable since we no longer need it.
    // Note that average count will always be at least one at this point, so there's no risk
    // dividing by zero.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    tempo = 0;
    for (i = 0; i < g_average_tempo_count; i++)
    {
        tempo += g_average_tempo[i];
    }
    tempo = tempo / g_average_tempo_count;

    return tempo;
}

/*====== A l i g n O u t p u t C l o c k S i g n a l ( ) ==============================================================
    Purpose     : 
    Author      : Harald Sabro
=====================================================================================================================*/

void AlignOutputClockSignal()
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Reset phase accumulator and wave table index for the base tempo.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    g_table_index = 0;
    g_phase_accumulator = 0;
}

/*====== S t a r t T e m p o C o u n t ( ) ============================================================================
    Purpose     : 
    Author      : Harald Sabro
=====================================================================================================================*/

void StartTempoCount()
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Sync the clock output and start the tempo counting.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    g_is_counting_tempo = true;
    g_tempo_count = 0;
}

/*====== E x i t T e m p o C o u n t ( ) ==============================================================================
    Purpose     : 
    Author      : Harald Sabro
=====================================================================================================================*/

void ExitTempoCount()
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Exit the tempo counting state and reset the tempo count.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    g_is_counting_tempo = false;
    g_tempo_count = 0;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Toggle off the "tap input registration active"-indicator LED in case it's on.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    DATA_REG_B |= (1 << TAP_INPUT_ACTIVE_PIN);
}

/*====== I S R ( T i m e r 1 ) ( ) ====================================================================================
    Purpose     : Timer1 overflow interrupt handler. This is where the output clock signals are drawn.
    Author      : Harald Sabro
    
    Frequency   : 31.25kHz
=====================================================================================================================*/

ISR(TIMER1_OVF_vect)
{
    uint8_t previous_base_table_index = g_table_index;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Increase the phase accumulator by a given amount based on the required output signal
    // frequency. Then use the high 8 bits (0-255) of the phase accumulator to identify what part
    // of the wave we're currently at.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    g_phase_accumulator += g_duty_cycle;
    g_table_index = (g_phase_accumulator & 0xff000000) >> 24;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Keep track of when there's an overflow in the base table index, i.e. the base tempo, has
    // just completed a full cycle.
    // Also keep track of when we're halfway through the cycle.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (previous_base_table_index > g_table_index)
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // As long as we're not currently in tempo counting mode (where the LED will be explicitly
        // set) toggle the LED states.
        //
        // The state of the LED outputs also doubles as base frequency clock output pulses (low one
        // complete base LFO cycle, then high one complete base LFO cycle) that can be used to
        // synchronize other controllers.
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        if (g_is_counting_tempo == false)
        {
            DATA_REG_B ^= (1 << CLOCK_OUT_PIN);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////
        // Note that, contrary to the regular clock pin, the 2x clock pin need to continue even
        // when we're currently counting tempo (see StartTempoCount() for more).
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        DATA_REG_B ^= (1 << CLOCK_OUT_DBL_PIN);
    }
    else if ((previous_base_table_index < 0x80) && (g_table_index >= 0x80))
    {
        DATA_REG_B ^= (1 << CLOCK_OUT_DBL_PIN);
    }
}

/*====== I S R ( T i m e r 0 ) ( ) ====================================================================================
    Purpose     : Timer0 compare interrupt handler. Triggers every TIMER0_FREQUENCY seconds.
    Author      : Harald Sabro
    
    Frequency   : 1kHz
    
    Note        : Some code borrowed from http://www.ganssle.com/debouncing-pt2.htm.
=====================================================================================================================*/

ISR(TIMER0_COMPA_vect)
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Start by counting tempo, if applicable.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (g_is_counting_tempo == true)
    {
        g_tempo_count++;        // Add a millisecond.

        ///////////////////////////////////////////////////////////////////////////////////////////
        // Make sure we don't exceed the maximum tempo length / minimum LFO frequency.
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        if (g_tempo_count > LFO_MIN_TAP_TEMPO)
        {
            ExitTempoCount();
        }
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Run each switch input through a debounce routine to make sure we get rid of any noise due to
    // the switch contacts bouncing.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
	DebounceSwitches();
	if ((g_switch_state_changed & (1 << TAP_INPUT_PIN)) && !(g_debounced_state & (1 << TAP_INPUT_PIN)))
	{
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Honor the input selection by only proceeding if tap input is selected as the source.
        ///////////////////////////////////////////////////////////////////////////////////////////

        if (g_input_selection == TAP_INPUT_SOURCE_ENABLED)
        {
            ///////////////////////////////////////////////////////////////////////////////////////
            // We've got a state change on the tap input switch pin going from open to closed
            // state, i.e. someone just stepped on the tap switch.
            ///////////////////////////////////////////////////////////////////////////////////////
            
            if (g_is_counting_tempo == false)
            {
                ///////////////////////////////////////////////////////////////////////////////////
                // Start counting tempo, making sure the output clock PWM signal is also reset.
                // Pull both the normal- and 2x clock pins low (start/on).
                ///////////////////////////////////////////////////////////////////////////////////

                StartTempoCount();
                AlignOutputClockSignal();

                DATA_REG_B &= ~((1 << CLOCK_OUT_PIN) | (1 << CLOCK_OUT_DBL_PIN));

                ///////////////////////////////////////////////////////////////////////////////////
                // Turn on the "tap input registration active"-indicator LED.
                ///////////////////////////////////////////////////////////////////////////////////
                
                DATA_REG_B &= ~(1 << TAP_INPUT_ACTIVE_PIN);
            }
            else
            {
                ///////////////////////////////////////////////////////////////////////////////////
                // Set the new base tempo now that we've got a millisecond count. Reset the output
                // clock PWM signal and end the tempo counting.
                // Pull the normal clock pin high (off), but the 2x pin low (on) since this signal
                // will have completed a full- rather than half a cycle.
                ///////////////////////////////////////////////////////////////////////////////////

                SetBaseTempo(g_tempo_count);
                ExitTempoCount();
                AlignOutputClockSignal();
                
                DATA_REG_B |= (1 << CLOCK_OUT_PIN);
                DATA_REG_B &= ~(1 << CLOCK_OUT_DBL_PIN);
            }
        }
	}
	else if ((g_switch_state_changed & (1 << SYNC_INPUT_PIN)) && !(g_debounced_state & (1 << SYNC_INPUT_PIN)))
	{
        ///////////////////////////////////////////////////////////////////////////////////////////
		// We've got a state change on the single sync input switch pin going from open to closed
        // state, i.e. someone just stepped on the single sync switch.
		///////////////////////////////////////////////////////////////////////////////////////////

        ///////////////////////////////////////////////////////////////////////////////////////////
        // Note that the single sync tap is only available when we're not running off an external
        // clock. It doesn't make sense to realign the base tempo only to have it undone on the
        // next clock input.
        ///////////////////////////////////////////////////////////////////////////////////////////

        if (g_input_selection != CLOCK_INPUT_SOURCE_ENABLED)
        {
            ///////////////////////////////////////////////////////////////////////////////////////
            // Reset the output clock PWM signal and end the tempo counting without setting a new
            // base tempo.
            // Pull the normal clock pin high (off), but the 2x pin low (on) since this signal will
            // have completed a full- rather than half a cycle.
            ///////////////////////////////////////////////////////////////////////////////////////

		    ExitTempoCount();
            AlignOutputClockSignal();
            
            DATA_REG_B |= (1 << CLOCK_OUT_PIN);
            DATA_REG_B &= ~(1 << CLOCK_OUT_DBL_PIN);
        }
        else
        {
            ///////////////////////////////////////////////////////////////////////////////////////
            // But if we are running off an external clock signal and that signal happens to also
            // be double speed, taping the single sync switch causes the next output clock pulse to
            // extend by 50%. This way you can select which one of the two incoming clock pulses
            // the single output clock pulse should align with.
            ///////////////////////////////////////////////////////////////////////////////////////

            if (g_input_clock_is_double_tempo == CONFIG_PIN_ENABLED)
            {
                if (g_is_counting_double_tempo == true)
                {
                    g_is_counting_double_tempo = false;
                }
                else
                {
                    g_is_counting_double_tempo = true;
                }
            }
        }
	}
}

/*====== I S R ( A D C ) ( ) ==========================================================================================
    Purpose     : Analog to Digital Converter interrupt handler.
    Author      : Harald Sabro
    
    Frequency   : 62.5kHz
=====================================================================================================================*/

ISR(ADC_vect)
{
    uint8_t channel = ADMUX & 0x1f;
    uint8_t sample_value = ADCH;
    uint8_t adjust_multiplier;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // We're running at 8MHz clock frequency with a prescaler of 128, i.e. 8MHz / 128 = 62.5kHz.
    // But we also take turns sampling two different channels meaning the actual sample frequency
    // is 62.5kHz / 2 = 31.25kHz. More than enough for our purposes.
    ///////////////////////////////////////////////////////////////////////////////////////////////
        
    if (channel == 0b101)   // Tempo adjust sample.
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Keep the tempo adjustment value updated at all times even though we might not be using
        // it. Allow a tiny bit of slack before changing since even a constant DC voltage will
        // fluctuate a bit.
        ///////////////////////////////////////////////////////////////////////////////////////////

        if ((sample_value > (g_tempo_adjustment + 1)) || (sample_value < (g_tempo_adjustment - 1)))
        {
            g_tempo_adjustment = sample_value;

            ///////////////////////////////////////////////////////////////////////////////////////
            // We've seen a change in the tempo adjustment. If the tempo adjustment feature is
            // enabled we proceed with recalculating tempo accordingly.
            ///////////////////////////////////////////////////////////////////////////////////////

            if (g_tempo_adjustment_is_active == CONFIG_PIN_ENABLED)
            {
                RecalculateTempo();
            }
        }
    }
    else if (channel == 0b110)   // Adjust range sample.
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Keep the adjustment range value updated at all times even though we might not be using
        // it. Allow a tiny bit of slack before changing since even a constant DC voltage will
        // fluctuate a bit.
        ///////////////////////////////////////////////////////////////////////////////////////////

        if ((sample_value > (g_adjustment_range + 5)) || (sample_value < (g_adjustment_range - 5)))
        {
            ///////////////////////////////////////////////////////////////////////////////////////
            // Keep the new adjustment range value so we have something to compare against, then
            // subdivide it to find the multiplier value.
            ///////////////////////////////////////////////////////////////////////////////////////

            g_adjustment_range = sample_value;
            adjust_multiplier = (sample_value / TEMPO_ADJUST_RANGE_DIVIDER) + TEMPO_ADJUST_MIN_MS_INTERVAL;

            if (g_adjustment_multiplier != adjust_multiplier)
            {
                g_adjustment_multiplier = adjust_multiplier;
                    
                ///////////////////////////////////////////////////////////////////////////////////
                // We've seen a change in the adjustment range. If the tempo adjustment feature is
                // enabled we proceed with recalculating tempo accordingly.
                ///////////////////////////////////////////////////////////////////////////////////

                if (g_tempo_adjustment_is_active == CONFIG_PIN_ENABLED)
                {
                    RecalculateTempo();
                }
            }
        }
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Switch the ADC input channel and initiate a new conversion.
    ///////////////////////////////////////////////////////////////////////////////////////////////
        
    ADMUX ^= (1 << MUX1) | (1 << MUX0);     // Toggle between MUX[2:0] = 0b101 and MUX[2:0] = 0b110.
    ADCSRA |= (1 << ADSC);
}

/*====== I S R ( P C I N T ) ( ) ======================================================================================
    Purpose     : Pin change interrupt handler. Reads input clock/sync pulses and config pin changes.
    Author      : Harald Sabro
=====================================================================================================================*/

ISR(PCINT_vect)
{
    uint8_t clock_input = (INPUT_PINS_A & (1 << CLOCK_INPUT_PIN)) >> 1;
    uint8_t tempo_adjust_enabled = (INPUT_PINS_A & (1 << ADJUST_ENABLED_PIN)) >> 3;
    uint8_t input_selection = (INPUT_PINS_A & (1 << INPUT_SRC_PIN)) >> 4;
    uint8_t clock_is_double_speed = (INPUT_PINS_A & (1 << CLOCK_SRC_IS_DBL_PIN)) >> 5;
    uint8_t tap_tempo_averaging = (INPUT_PINS_B & (1 << TAP_AVERAGING_PIN)) >> 6;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Check for a change on the input selection pin.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    if (input_selection != g_input_selection)
    {
        g_input_selection = input_selection;

        ///////////////////////////////////////////////////////////////////////////////////////////
        // We have a change in the input source selection. If we were in the middle of counting
        // tempo we need to stop doing that. This is specially important if we're transitioning
        // away from using the clock input as source since there's a 50% chance we'll get stuck
        // inside a tempo counting state.
        ///////////////////////////////////////////////////////////////////////////////////////////

        ExitTempoCount();

        ///////////////////////////////////////////////////////////////////////////////////////////
        // We also wipe any previously stored tempos in case of averaging.
        ///////////////////////////////////////////////////////////////////////////////////////////

        g_average_tempo_count = 0;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Check for a change on the tempo adjust enable pin.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    if (tempo_adjust_enabled != g_tempo_adjustment_is_active)
    {
        g_tempo_adjustment_is_active = tempo_adjust_enabled;

        RecalculateTempo();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Check for a change on the tap tempo averaging enable pin.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    if (tap_tempo_averaging != g_tap_tempo_averaging)
    {
        g_tap_tempo_averaging = tap_tempo_averaging;

        ///////////////////////////////////////////////////////////////////////////////////////////
        // Invalidate any accumulated tap tempo readings when averaging is disabled.
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        if (g_tap_tempo_averaging == CONFIG_PIN_DISABLED)
        {
            g_average_tempo_count = 0;
        }
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Check for a change on the double speed clock input enable pin.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    if (clock_is_double_speed != g_input_clock_is_double_tempo)
    {
        g_input_clock_is_double_tempo = clock_is_double_speed;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Check for a change on the clock input pin.
    ///////////////////////////////////////////////////////////////////////////////////////////////

    if (clock_input != g_clock_input)
    {
        g_clock_input = clock_input;

        ///////////////////////////////////////////////////////////////////////////////////////////
        // Honor the input selection by only proceeding if clock input is selected the source.
        ///////////////////////////////////////////////////////////////////////////////////////////

        if (g_input_selection == CLOCK_INPUT_SOURCE_ENABLED)
        {
            ///////////////////////////////////////////////////////////////////////////////////////
            // Determine if we're running off a normal or a 2x speed clock source.
            ///////////////////////////////////////////////////////////////////////////////////////

            if (g_input_clock_is_double_tempo == CONFIG_PIN_ENABLED)
            {
                ///////////////////////////////////////////////////////////////////////////////////
                // We're running off a 2x speed clock source and need to take this into account
                // when counting tempo and generating our own output clock signal.
                ///////////////////////////////////////////////////////////////////////////////////

                if (!(INPUT_PINS_A & (1 << CLOCK_INPUT_PIN)))
                {
                    ///////////////////////////////////////////////////////////////////////////////
                    // Rather than counting tempo from start edge to stop edge (half a clock cycle
                    // really) like we normally do, we count all the way to the next start edge.
                    // This means we have to locally keep track of how to treat each start edge.
                    ///////////////////////////////////////////////////////////////////////////////

                    if (g_is_counting_double_tempo == true)
                    {
                        ///////////////////////////////////////////////////////////////////////////
                        // See Timer0 interrupt handler.
                        ///////////////////////////////////////////////////////////////////////////

                        SetBaseTempo(g_tempo_count);
                        ExitTempoCount();
                        AlignOutputClockSignal();
                        
                        DATA_REG_B |= (1 << CLOCK_OUT_PIN);
                        DATA_REG_B &= ~(1 << CLOCK_OUT_DBL_PIN);

                        g_is_counting_double_tempo = false;
                    }
                    else
                    {
                        ///////////////////////////////////////////////////////////////////////////
                        // See Timer0 interrupt handler.
                        ///////////////////////////////////////////////////////////////////////////

                        StartTempoCount();
                        AlignOutputClockSignal();

                        DATA_REG_B &= ~((1 << CLOCK_OUT_PIN) | (1 << CLOCK_OUT_DBL_PIN));

                        g_is_counting_double_tempo = true;
                    }
                }
            }
            else
            {
                ///////////////////////////////////////////////////////////////////////////////////
                // Detect whether this is a falling or rising edge, and start or stop the tempo
                // counting accordingly.
                ///////////////////////////////////////////////////////////////////////////////////
                
                if (INPUT_PINS_A & (1 << CLOCK_INPUT_PIN))
                {
                    ///////////////////////////////////////////////////////////////////////////////
                    // See Timer0 interrupt handler.
                    ///////////////////////////////////////////////////////////////////////////////

                    SetBaseTempo(g_tempo_count);
                    ExitTempoCount();
                    AlignOutputClockSignal();
                    
                    DATA_REG_B |= (1 << CLOCK_OUT_PIN);
                    DATA_REG_B &= ~(1 << CLOCK_OUT_DBL_PIN);
                }
                else
                {
                    ///////////////////////////////////////////////////////////////////////////////
                    // See Timer0 interrupt handler.
                    ///////////////////////////////////////////////////////////////////////////////

                    StartTempoCount();
                    AlignOutputClockSignal();

                    DATA_REG_B &= ~((1 << CLOCK_OUT_PIN) | (1 << CLOCK_OUT_DBL_PIN));
                }
            }
        }
    }
}

/*====== End of file ================================================================================================*/